#define GLOBAL_GETSQLCODE_DATA 1
/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : getSQLCode.c                                             */
/* Output name     : getSQLCode.dll                                           */
/*                   (NOTE: must be renamed to getSQLCode.lil and moved to    */
/*                   <install dir>\bin for use with MQSI V2.0 broker          */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\include\plugin        */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : This plug-in node extracts the SQLCODE value from the    */
/*                   message exception list.                                  */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Jul 2001                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include "getSQLCode.h" /* Specific header for this dll */                       
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "plugin.h"   /* common plugin node functions */
#include <BipCni.h>	  /* Function prototypes for Plug-in Node functions     */

/* local function prototypes */


/******************************************************************************/
/* Node Implementation Functions                                              */
/*                                                                            */
/* These are mandatory and MUST be implemented by the programmer.             */
/*                                                                            */
/* createNodeContext                                                          */
/* deleteNodeContext                                                          */
/* evaluate                                                                   */
/* getAttribute                                                               */
/* getAttributeName                                                           */
/* setAttribute                                                               */
/* bipGetMessageflowNodeFactory                                               */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/* createNodeContext                                                          */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  factoryObject (input) - address of factory object that owns the node      */
/*                          being created.                                    */
/*  nodeName      (input) - name of node being created.                       */
/*  nodeObject    (input) - address of the node object that has just been     */
/*                          created.                                          */
/* Returns:                                                                   */
/*                                                                            */
/*  Pointer to the local context.                                             */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Creates any plug-in context for an instance of the node object. This is    */
/* invoked by the broker whenever a node instance is constructed. Nodes are   */
/* constructed whenever a messageflow containing the node is deployed by the  */
/* broker.                                                                    */
/*                                                                            */
/* This function has 4 responsibilities:                                      */
/*                                                                            */
/* 1. Optionally verify the name of the node is supported by the factory.     */
/* 2. Allocate node instance specific data areas. e.g. For attribute data or  */
/*    terminals.                                                              */
/* 3. Perform any additional resource acquisition or initialisation that is   */
/*    required for the processing of the node.                                */
/* 4. Return the address of the context to the calling function.              */
/*                                                                            */
/******************************************************************************/
CciContext *createNodeContext(CciFactory *factoryObject, 
                              CciChar *nodeName, 
                              CciNode *nodeObject)
{
    int rc;
	static char *functionName="createNodeContext()";
    context_t *contextP = NULL;
#if defined(POST_201_VERSION)
	CNI_BROKER_INFO_ST brokerInfo={CNI_BROKER_INFO_ST_DEFAULT};
#else
    CNI_BROKER_INFO_ST brokerInfo;
#endif
	CciChar messageFlowNameUnicode[MAX_STRING_SIZE + 1];
	char messageFlowNameMB[MAX_STRING_SIZE + 1];
	messageFlow_t *mfP,
	              defaultMsgFlow = {MESSAGE_FLOW_DEFAULT},
		          *messageFlowAnchorP;
	deployedGetSQLCode_t *deployedGetSQLCodeP;
	char msgBuffer[200],
		 nodeNameMB[100];

    TRACE_ENTRY("createNodeContext", ("factoryObject=%.8x, nodeName=%S, "
		                              "nodeObject=%.8x",
									  factoryObject,
									  nodeName,
									  nodeObject));

	/* initialise the broker info structure */
	ALLOCATE_STRING_STORAGE(brokerInfo.brokerName, 300);
	ALLOCATE_STRING_STORAGE(brokerInfo.executionGroupName, 300);
	ALLOCATE_STRING_STORAGE(brokerInfo.messageFlowName, 300);
	ALLOCATE_STRING_STORAGE(brokerInfo.queueManagerName, 300);

	/* get the name of our queue manager */
    TRACE_MID(TRACE_INFO, ("About to call cniGetBrokerInfo"));

	cniGetBrokerInfo(0,           /* retcode does not appear to be set */
			         nodeObject,
					 &brokerInfo);

	UCsprintf(messageFlowNameMB, "%S", brokerInfo.messageFlowName.buffer);
	UCcpy(messageFlowNameUnicode, brokerInfo.messageFlowName.buffer);
    TRACE_MID(TRACE_INFO, ("Extracted messageFlowName is '%s'",
			               messageFlowNameMB));

	/* Is this the first getSQLCode for this Message Flow */
	threadMutexLock(initMutex);
	mfP = findMessageFlow(messageFlowNameUnicode, &messageFlowAnchor);

	if (!mfP)
	{
		/* create new message flow entry */
		mfP = (messageFlow_t *)malloc(sizeof(messageFlow_t));
		memcpy(mfP, &defaultMsgFlow, sizeof(messageFlow_t));
		mfP->name = UCdup(messageFlowNameUnicode);

		/* add to message flow chain */
		messageFlowAnchorP = &messageFlowAnchor;
		INSERT_AFTER_IN_LINKED_LIST(messageFlowAnchorP, mfP);
	}
	threadMutexUnlock(initMutex);

    TRACE_MID(TRACE_INFO, ("Current messageFlow entry at %.8x", mfP));

	if (!brokerInfoObtained)
	{
		/* copy the process specific information, both as single and Unicode strings */
		UCsprintf(brokerName, "%S", brokerInfo.brokerName.buffer);
		UCcpy(brokerNameUC, brokerInfo.brokerName.buffer);
        TRACE_MID(TRACE_INFO, ("Extracted brokerName is '%s'",
			                   brokerName));

		UCsprintf(executionGroupName, "%S", brokerInfo.executionGroupName.buffer);
		UCcpy(executionGroupNameUC, brokerInfo.executionGroupName.buffer);
        TRACE_MID(TRACE_INFO, ("Extracted executionGroupName is '%s'",
			                   executionGroupName));

		UCsprintf(queueManagerName, "%S", brokerInfo.queueManagerName.buffer);
		UCcpy(queueManagerNameUC, brokerInfo.queueManagerName.buffer);
        TRACE_MID(TRACE_INFO, ("Extracted queueManagerName is '%s'",
			                   queueManagerName));

		brokerInfoObtained = TRUE;
	}

	FREE_STRING_STORAGE(brokerInfo.brokerName);
	FREE_STRING_STORAGE(brokerInfo.executionGroupName);
	FREE_STRING_STORAGE(brokerInfo.messageFlowName);
	FREE_STRING_STORAGE(brokerInfo.queueManagerName);

    /* Allocate a pointer to the local context */
    contextP = malloc(sizeof(context_t));

    /* If malloc was successful */
    if (!contextP)
	{
        /* Error handling if malloc failed */
        TRACE_RETURN_TYPE(RSN_022, void *, contextP);
	}

    TRACE_MID(TRACE_INFO, ("Context structure allocated at %.8x", contextP));

    /* initialise returned storage for context anchor */
    memset (contextP, 0, sizeof(context_t)); /* Clear the context area */
    contextP->nodeObject = nodeObject;
	contextP->messageFlowName = mfP->name;

    TRACE_MID(TRACE_INFO, ("<===========================>"));
    TRACE_MID(TRACE_INFO, ("%S node current", nodeName));
    TRACE_MID(TRACE_INFO, ("<===========================>"));

    /* Verify the name of the node is supported by the factory */
    if (!UCcmp(nodeName, C.GET_SQL_CODE_NODE))
	{
		/* getSQLCode save node */
		contextP->nodeType = GET_SQL_CODE;
        contextP->nodeName = UCdup(nodeName);

		/* Build attribute list for the node */
		insertAttributeTableEntry(contextP, nodeName, TRUE,
			                      C.ATTR_COPY_VALUE_TO, CCI_VALUE_TYPE_BYTEARRAY);

		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_LABEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_USER_TRACE_LEVEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_TRACE_LEVEL, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_USER_TRACE_FILTER, CCI_VALUE_TYPE_BYTEARRAY);
		insertAttributeTableEntry(contextP, nodeName, FALSE,
			                      C.ATTR_TRACE_FILTER, CCI_VALUE_TYPE_BYTEARRAY);

		/* this node has a single input and 3 output terminals */
		insertInputTerminalListEntry(contextP, C.TERM_IN);

		insertOutputTerminalListEntry(contextP, C.TERM_OUT);
		insertOutputTerminalListEntry(contextP, C.TERM_FAILURE);
		insertOutputTerminalListEntry(contextP, C.TERM_NOT_FOUND);

		/* create a new deployed getSQLCode control block */
		deployedGetSQLCodeP = (deployedGetSQLCode_t *)
			                         malloc(sizeof(deployedGetSQLCode_t));
		memset(deployedGetSQLCodeP, 0, sizeof(deployedGetSQLCode_t));

		/* lock the getSQLCode mutex and add to global chain */
		threadMutexLock(initMutex);
		INSERT_AFTER_IN_LINKED_LIST(getSQLCodeAnchorP,
					                deployedGetSQLCodeP);
		contextP->nodeDataP = deployedGetSQLCodeP;

		/* increment the global node counter */
		numGetSQLCodeNodes++;
	    TRACE_MID(TRACE_INFO, ("Global count of GetSQLCode nodes incremented to %d",
			                   numGetSQLCodeNodes));
		threadMutexUnlock(initMutex);

		/* As of MQSI V2.0.2, an attribute for which the user has accepted the default, is
		 * no longer included in the deployment message.  We therefore need to initialise
		 * all attributes to their default values within the plug-in.
		 * The exception to this is a value for which no 'value=xxx' label is specified and 
		 * for which 'valueMandatory=true' has been specified.  In this case, the control 
		 * centre will force the user to enter a value into the field, and this will be 
		 * included in the deployment message.  Of course, we must still check for invalid
		 * values such as blanks.
		 */
//	    TRACE_MID(TRACE_INFO, ("Setting default values"));
//		setAttribute(contextP,
//		             C.ATTR_COPY_VALUE_TO,
//                   C.ATTR_COPY_VALUE_TO_DEFAULT_VALUE);
	}
	else
	{
	    /* Error handling if name of node is not supported by factory */
        TRACE_MID(TRACE_ERROR, ("Unrecognised node"));
		/* throw an exception here...*/
		UCsprintf(msgBuffer, "Unrecognised node ('%S') passed to createContext()",
			               contextP->nodeName);
		UCsprintf(nodeNameMB, "'%S'", contextP->nodeName);
		cciThrowException(&rc, 
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_UNRECOGNISED_NODE_CREATE_CONTEXT,  /* messageNumber */ 
			              msgBuffer,              /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
						  nodeNameMB,
			              (char *)0);
	    free(contextP);
	}

    /* Return the address of the context to the calling function */
    TRACE_RETURN_TYPE(RSN_008, void *, contextP);
}

/******************************************************************************/
/* deleteNodeContext                                                          */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input)  - pointer to the current context                   */
/*                          created.                                          */
/* Returns:                                                                   */
/*                                                                            */
/*  Nothing.                                                                  */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Called whenever an instance of a node object is destroyed.  Any resources  */
/* aquired by the node e.g. context or other resources, should be released    */
/* at this time.                                                              */
/*                                                                            */
/******************************************************************************/
void deleteNodeContext(CciContext *dataP)
{
	context_t *contextP = (context_t *)dataP;
	int rc;
    static char *functionName="deleteNodeContext()";
	deployedGetSQLCode_t *deployedGetSQLCodeP;
	char infoMessage[1000],
		 *nodeNameP,
	     mbLabelString[MAX_STRING_SIZE + 1];

    TRACE_ENTRY("deleteNodeContext", ("dataP=%.8x", dataP));

	/* lock mutex just to be sure there's no race */
	TRACE_MID(TRACE_INFO, ("About to lock initMutex"));
	threadMutexLock(initMutex);

	/* delete deployed node structure pointed to by this context */
	if (contextP->nodeType == GET_SQL_CODE)
	{
		deployedGetSQLCodeP = (deployedGetSQLCode_t *)contextP->nodeDataP;
		TRACE_MID(TRACE_INFO, ("Processing deployed getSQLCode at %.8x", deployedGetSQLCodeP));

		/* remove deployed getSQLCode from global chain */
		REMOVE_FROM_LINKED_LIST(deployedGetSQLCodeP);

		/* free any linked resources */
		if (deployedGetSQLCodeP->copyValueToLocation)
		{
			TRACE_MID(TRACE_INFO, ("Freeing deployed getSQLCode copy location name"));
			free(deployedGetSQLCodeP->copyValueToLocation);
		}

		/* finally, free the deployed node control block itself */
		free(deployedGetSQLCodeP);
		TRACE_MID(TRACE_INFO, ("Freeing deployed getSQLCode at %.8x", deployedGetSQLCodeP));

		/* decrement number of create nodes */
		numGetSQLCodeNodes--;
		TRACE_MID(TRACE_INFO, ("Number of getSQLCode nodes decremented to %d",
			                   numGetSQLCodeNodes));
	}

	TRACE_MID(TRACE_INFO, ("About to unlock initMutex"));
	threadMutexUnlock(initMutex);

	/* clear the attributes associated with this context */
    deleteAttributeTableEntries(contextP);

	/* clear the terminals associated with this context */
    deleteAllTerminalListEntries(contextP);

	/* 
	 * Finally, free up the context structure, but first create multibyte strings for
	 * inclusion in log message.
	 */
    UCToStr(contextP->nodeLabel,
		    mbLabelString,
		    sizeof(mbLabelString),
		    NULL);

	if (contextP->nodeType == GET_SQL_CODE)
	{
		nodeNameP = GET_SQL_CODE_NODE_MB;
	}

	free(contextP->nodeName);
	free(contextP->nodeLabel);
    free(contextP);

	sprintf(infoMessage, "Successfully deleted node context for %s (%s)", 
		                  nodeNameP, mbLabelString);
	TRACE_MID(TRACE_INFO, ("%s", infoMessage));

    cciLog(&rc,                       /* returnCode */
	       CCI_LOG_INFORMATION,       /* type */
		   __FILE__,                  /* file */
		   __LINE__,                  /* line */
		   functionName,              /* function */
		   C.PLUGIN_MESSAGE_CATALOGUE, /* messageSource */
           MJB_NODE_CONTEXT_DELETED,  /* messageNumber */ 
		   infoMessage, /* traceText */
#if !defined(WIN32)
		   brokerName,
#endif
           nodeNameP,
		   mbLabelString,
		   (char *)0);

	TRACE_RETURN_VOID(RSN_009);
}


/******************************************************************************/
/* setAttribute                                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context        (input)  - pointer to the current context                  */
/*  attributeName  (input)  - attribute name for which value is to be set     */
/*  attributeValue (input) -  value to be set for the attribute               */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Sets the value of an attribute on a specific node instance. It is invoked  */
/* by the broker when a configuration request is received that attempts to    */
/* set the value of a node attribute. The broker will have verified that the  */
/* attribute name is valid for the node.                                      */
/*                                                                            */
/* This function has 3 responsibilities:                                      */
/*                                                                            */
/* 1. Verify value of attribute is correctly specified, otherwise a           */
/*    configuration exception should be thrown.                               */
/* 2. Store the attribute value in the context.                               */
/* 3. Throw a configuration exception if errors occur.                        */
/*                                                                            */
/******************************************************************************/
int setAttribute(CciContext *dataP,
				 CciChar *attributeName,
				 CciChar *attributeValue)
{
	int rc;
	context_t *contextP = (context_t *)dataP;

    TRACE_ENTRY("setAttribute", ("dataP=%.8x, attributeName=%S, "
		                         "attributeValue='%S'",
								 dataP, attributeName, attributeValue));

	/* call common plugin function */
	rc = commonSetAttribute(contextP,
		                    attributeName,
						    attributeValue);

	if (rc != CCI_SUCCESS)
	{
        TRACE_RETURN_TYPE(RSN_011, int, rc);
	}

	/* special processing for getSQLCode attributes */
    rc = processAttribute(contextP,
		                  attributeName,
		                  attributeValue);
	if (rc != TT_SUCCESS)
	{
	    TRACE_RETURN_TYPE(RSN_012, int, CCI_EXCEPTION_CONFIGURATION);
	}
    TRACE_RETURN_TYPE(RSN_012, int, TT_SUCCESS);
}


/******************************************************************************/
/* processAttribute                                                           */
/******************************************************************************/
int processAttribute(IN context_t *contextP,
				     IN CciChar *attributeName,
				     IN CciChar *attributeValue)
{
	deployedGetSQLCode_t *deployedGetSQLCodeP;
    CciChar copyString[MAX_STRING_SIZE + 1];
	int copyStringLength;
	char msgBuffer[2000],
		 nodeName[1000],
		 propertyName[100];

	static char *functionName = "processAttribute";

    TRACE_ENTRY("processAttribute", ("contextP=%.8x, attributeName=%S, "
		                             "attributeValue='%S'",
								     contextP, attributeName, attributeValue));

	/* 
	 * During message flow initialisation, it's unlikely that this code
	 * will be run concurrently.  However, just to be sure, we'll 
	 * serialise the code through this section.
	 */
	TRACE_MID(TRACE_INFO, ("About to lock initMutex"));
	threadMutexLock(initMutex);

	if (contextP->nodeType == GET_SQL_CODE)
	{
		deployedGetSQLCodeP = (deployedGetSQLCode_t *)(contextP->nodeDataP);

		/* initialise the deployed getSQLCode structure */

		/* process the MessageData property */
		if (!UCcmp(attributeName, C.ATTR_COPY_VALUE_TO))
		{
			copyStringLength = sizeof(copyString);
			if (getStringAttribute(contextP, 
					               C.ATTR_COPY_VALUE_TO,
								   &copyStringLength,
					               copyString) == TT_SUCCESS)
			{
				/* 
				 * The string value must begin with 'Root.' - detect the following invalid values
				 *  - no value specified (not sure how this might happen)
				 *  - doesn't start with Root.
				 *  - contains only Root.XML 
				 *  - contains only Root.XML.
				 */
			    if (!UClen(copyString) ||
					UClen(copyString) <= 5 ||
					UCncmp(copyString, C.ROOT_DOT, UClen(C.ROOT_DOT)) ||
					!UCcmp(copyString, C.ROOT_DOT_XML) ||
					!UCcmp(copyString, C.ROOT_DOT_XML_DOT))
				{
					TRACE_MID(TRACE_INFO, ("Invalid value specified for %s attribute", 
						                   ATTR_COPY_VALUE_TO_MB));
					// generate exception

					UCsprintf(msgBuffer, "The syntax element name specified in the %S "
						               "property of node %S(%S) is invalid. "
									   "Valid names must begin with 'Root.' or 'Root.XML.'",
									   C.ATTR_COPY_VALUE_TO,
									   contextP->nodeName,
									   contextP->nodeLabel ? contextP->nodeLabel : C.NULL_LITERAL);

					TRACE_MID(TRACE_ERROR, ("%s", msgBuffer));

					UCsprintf(propertyName, "%S", C.ATTR_COPY_VALUE_TO);

					UCsprintf(nodeName, "%S(%S)", contextP->nodeName,
							contextP->nodeLabel ? contextP->nodeLabel : C.NULL_LITERAL);

					threadMutexUnlock(initMutex);

					TraceSetIndentationValue(0);

					cciThrowException(0, 
						   CCI_CONFIGURATION_EXCEPTION,
						   __FILE__,
						   __LINE__,
						   functionName,
						   C.PLUGIN_MESSAGE_CATALOGUE,
						   MJB_INVALID_COPY_VALUE_TO_SYNTAX_ELEMENT_NAME,  /* messageNumber */ 
						   msgBuffer,              /* traceText */
#if !defined(WIN32)
						   brokerName,
#endif
						   propertyName,
						   nodeName,
						   (char *)0);
				}

				TRACE_MID(TRACE_INFO, ("Copying syntax element path '%S' to deployed getSQLCode",
					                    copyString));

				/* copy string into deployed getSQLCode, freeing any default value first */
				if (deployedGetSQLCodeP->copyValueToLocation)
				{
					free(deployedGetSQLCodeP->copyValueToLocation);
				}
				deployedGetSQLCodeP->copyValueToLocation = UCdup(copyString);
			}
		}
	}

	TRACE_MID(TRACE_INFO, ("About to unlock initMutex"));
	threadMutexUnlock(initMutex);
								  
    TRACE_RETURN(RSN_013, TT_SUCCESS);
}

/******************************************************************************/
/* evaluate                                                                   */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Performs node processing. This function contains the main processing logic */
/* of the message flow node. It is invoked by the broker when a message is    */
/* received on one of the input terminals of an instance of a node object.    */
/*                                                                            */
/* This function has 4 responsibilities:                                      */
/*                                                                            */
/* 1. Process the message in accordance with the values of any attribute on   */
/*    the node instance.                                                      */
/* 2. Process the message based on content if required.                       */
/* 3. Propagate the message to the appropriate output terminal                */
/* 4. Throw an exception if errors occur.                                     */
/*                                                                            */
/******************************************************************************/
void evaluate(CciContext *dataP,
		      CciMessage *destinationList,
		      CciMessage *exceptionList,
		      CciMessage *message)
{
	context_t *contextP;
	int rc;
	static char *functionName = "evaluate";
	char msgBuffer[200],
		 nodeName[100];

    TRACE_ENTRY("evaluate", ("dataP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 dataP, destinationList,
							 exceptionList, message));

	contextP = (context_t *)dataP;

	/* determine which function to call based on node name */
    TRACE_MID(TRACE_INFO, ("nodeName='%S'", contextP->nodeName));
	if (!UCcmp(contextP->nodeName, C.GET_SQL_CODE_NODE))
	{
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"));
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>>  GetSQLCode  >>>>>>>>>>>>"));
        TRACE_MID(TRACE_INFO, (">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"));
		rc = evaluateGetSQLCode(contextP,
			                    destinationList,
						        exceptionList,
						        message);
	}
	else
	{
        TRACE_MID(TRACE_ERROR, ("Unrecognised node"));
		/* throw an exception here...*/
		UCsprintf(msgBuffer, "Unrecognised node ('%S') passed to evaluate()",
			               contextP->nodeName);
		UCsprintf(nodeName, "'%S'", contextP->nodeName);
		cciThrowException(&rc, 
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_UNRECOGNISED_NODE,  /* messageNumber */ 
			              msgBuffer,              /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
						  nodeName,
			              (char *)0);
	}

    TRACE_RETURN_VOID(RSN_016);
}

/******************************************************************************/
/* bipGetMessageflowNodeFactory                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  Address of factory object.                                                */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* This function is called when the message broker loads the "lil" during the */
/* initialization of the broker engine                                        */
/*                                                                            */
/* This function has 5 responsibilities:                                      */
/*                                                                            */
/* 1. create the node factory (or factories) which the "lil" supports, using  */
/*    the createNodeFactory() function, saving the returned pointer to the    */
/*    node factory for use at runtime.                                        */
/* 2. define each node supported by the (or each) factory) and pass a pointer */
/*    to a function table which contains pointers to the node implementation  */
/*    functions.                                                              */
/* 3. define the name of each attribute supported by each node, including a   */
/*    pointer to a validation function.                                       */
/* 4. define input and output terminals supported by each node.               */
/* 5. return the pointer to the factory, unless an error occurred, in which   */
/*    case zero should be returned.                                           */    
/*                                                                            */
/******************************************************************************/
#if defined(WIN32)
    #include <process.h>
#elif defined(_AIX) || defined(_SOLARIS) 
    #include <unistd.h>
#endif
#if defined(WIN32)
void __declspec(dllexport) *bipGetMessageflowNodeFactory()
#else
CciFactory LilFactoryExportPrefix * LilFactoryExportSuffix bipGetMessageflowNodeFactory()
#endif
{
    static char *functionName="bipGetMessageflowNodeFactory()";
    void* factoryObject = NULL;
    CciChar *factoryName = C.GETSQLCODE_PLUGIN_FACTORY;
#if defined(POST_201_VERSION)
	CCI_EXCEPTION_ST exception_st={CCI_EXCEPTION_ST_DEFAULT};
#else
    CCI_EXCEPTION_ST exception_st;
#endif
    int rc;
#if defined(POST_201_VERSION)
	static CNI_VFT vftable={CNI_VFT_DEFAULT};
#else
    static CNI_VFT vftable;
#endif
	bool_t resourceCreated = TRUE;
	char *resourceName,
		 msgBuffer[200];

    TRACE_ENTRY("bipGetMessageflowNodeFactory", ("void"));
    currentPID = getpid();
    TRACE_MID(TRACE_INFO, ("<------------------------------------------------------>"));
    TRACE_MID(TRACE_INFO, ("<**** Start of getSQLCode deployment (pid=%d) ****>",
		                   currentPID));
    TRACE_MID(TRACE_INFO, ("<------------------------------------------------------>"));
	/* create the global syncronisation resources we need here */

	/* create required mutexes */
	if (resourceCreated)
	{
        TRACE_MID(TRACE_INFO, ("About to create initMutex"));
		if (threadCreateMutex(FALSE,  /* create in unlocked state */
							  &initMutex) != THREAD_OK)
		{
			resourceName = "initMutex";
			resourceCreated = FALSE;
		}
	}

	/* process any errors that have occurred */
	if (!resourceCreated)
	{
		sprintf(msgBuffer, "GetSQLCode Plugin Factory failed to create resource %s",
						   resourceName);
		/* generate an exception here... */
		cciThrowException(&rc, 
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_RESOURCE_CREATION_FAILED,  /* messageNumber */ 
			              msgBuffer, /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
			              "GetSQLCodePlugInFactory",
						  resourceName,
			              (char *)0);
	}


    /* Setup virtual function table with addresses of implementation functions */
#if !defined(POST_201_VERSION)
    memset(&vftable, 0, sizeof(vftable));
#endif
    vftable.iFpCreateNodeContext     = createNodeContext;
    vftable.iFpDeleteNodeContext     = deleteNodeContext;
    vftable.iFpGetAttributeName      = getAttributeName;
    vftable.iFpSetAttribute          = setAttribute;
    vftable.iFpGetAttribute          = getAttribute;
    vftable.iFpEvaluate              = evaluate;


    /* Create the Node Factory for this plug-in */
    factoryObject = cniCreateNodeFactory(&rc, factoryName);
    if (factoryObject == CCI_NULL_ADDR)
	{
        if (rc == CCI_EXCEPTION)
		{
            cciGetLastExceptionData(&rc, 
				                    &exception_st);
            /* Any local error handling may go here */
            cciRethrowLastException(0);
		}
        /* Any further local error handling can go here */
        return(factoryObject);
	}
    /* Define the GetSQLCodeNode node supported by our factory */

	/*
	 * Define the GetSQLCodeNode node
	 */
    cniDefineNodeClass(&rc, 
			           factoryObject,
					   C.GET_SQL_CODE_NODE,
					   &vftable); 
    if (rc != CCI_SUCCESS)
	{
        TRACE_MID(TRACE_ERROR, ("Failed to create GetSQLCodeNode node "
			                    "(cniDefineNodeClass rc=%d)", rc));
		/* generate an exception here... */
		cciThrowException(&rc,
		                  CCI_FATAL_EXCEPTION,
					      __FILE__,
					      __LINE__,
					      functionName,
					      C.PLUGIN_MESSAGE_CATALOGUE,
			              MJB_NODE_CLASS_CREATION_FAILED,  /* messageNumber */ 
			              "GetSQLCode Plugin Factory failed to create GetSQLCodeNode node", /* traceText */
#if !defined(WIN32)
						  brokerName,
#endif
			              "GetSQLCodePlugInFactory",
			              GET_SQL_CODE_NODE_MB,
			              (char *)0);
	}

    /* Log a success message */
    TRACE_MID(TRACE_INFO, ("About to issue cciLog success message"));
    cciLog(&rc,                       /* returnCode */
	       CCI_LOG_INFORMATION,       /* type */
		   __FILE__,                  /* file */
		   __LINE__,                  /* line */
		   functionName,              /* function */
		   C.PLUGIN_MESSAGE_CATALOGUE,  /* messageSource */
#if defined(WIN32)
           MJB_NODE_FACTORY_CREATED,  /* messageNumber */ 
#else
           MJB_NODE_FACTORY_CREATED_UNIX,  /* messageNumber */ 
#endif
		   "GetSQLCode Plugin Factory loading", /* traceText */
		   "GetSQLCodePlugInFactory",
		   (char *)0);

    /* Return address of this factory object to the broker */
    TRACE_RETURN_TYPE(RSN_021, void *, factoryObject);
}


/* local functions */

